""" Generator functions to evaluate trial and test functions

NB: Basis evaluation is only needed for the trial function argument in jacobians, as the
multiplication with the test function is part of the sum factorization kernel.
"""
import itertools

from dune.perftool.generation import (backend,
                                      domain,
                                      get_backend,
                                      get_counted_variable,
                                      get_counter,
                                      get_global_context_value,
                                      iname,
                                      instruction,
                                      kernel_cached,
                                      temporary_variable,
                                      )
from dune.perftool.sumfact.tabulation import (basis_functions_per_direction,
                                              construct_basis_matrix_sequence,
                                              BasisTabulationMatrix,
                                              quadrature_points_per_direction,
                                              PolynomialLookup,
                                              name_polynomials,
                                              polynomial_degree,
                                              )
from dune.perftool.sumfact.quadrature import quadrature_inames
from dune.perftool.sumfact.switch import (get_facedir,
                                          get_facemod,
                                          )
from dune.perftool.pdelab.argument import name_coefficientcontainer
from dune.perftool.pdelab.geometry import (local_dimension,
                                           world_dimension,
                                           )
from dune.perftool.loopy.buffer import initialize_buffer, get_buffer_temporary
from dune.perftool.sumfact.symbolic import SumfactKernel, SumfactKernelInputBase
from dune.perftool.options import get_option
from dune.perftool.pdelab.driver import FEM_name_mangling
from dune.perftool.pdelab.restriction import restricted_name
from dune.perftool.pdelab.spaces import name_lfs, name_lfs_bound, name_leaf_lfs
from dune.perftool.tools import maybe_wrap_subscript
from dune.perftool.pdelab.basis import shape_as_pymbolic
from dune.perftool.sumfact.accumulation import sumfact_iname

from ufl import MixedElement, VectorElement, TensorElement, TensorProductElement

from pytools import product, ImmutableRecord

from loopy.match import Writes

import pymbolic.primitives as prim


class LFSSumfactKernelInput(SumfactKernelInputBase, ImmutableRecord):
    def __init__(self,
                 coeff_func=None,
                 element=None,
                 element_index=0,
                 restriction=0,
                 ):
        ImmutableRecord.__init__(self,
                                 coeff_func=coeff_func,
                                 element=element,
                                 element_index=element_index,
                                 restriction=restriction,
                                 )

    def realize(self, sf, index, insn_dep):
        lfs = name_lfs(self.element, self.restriction, self.element_index)
        basisiname = sumfact_iname(name_lfs_bound(lfs), "basis")
        container = self.coeff_func(self.restriction)
        from dune.perftool.pdelab.argument import pymbolic_coefficient as pc
        coeff = pc(container, lfs, basisiname)

        # Get the input temporary!
        name = get_buffer_temporary(sf.buffer,
                                    shape=(product(mat.basis_size for mat in sf.matrix_sequence), sf.vector_width),
                                    name="input_{}".format(sf.buffer)
                                    )

        assignee = prim.Subscript(prim.Variable(name),
                                  (prim.Variable(basisiname),) + (index,))
        instruction(assignee=assignee,
                    expression=coeff,
                    depends_on=sf.insn_dep.union(insn_dep),
                    tags=frozenset({"sumfact_stage{}".format(sf.stage)}),
                    )

    @property
    def direct_input(self):
        if get_option("fastdg"):
            return self.coeff_func(self.restriction)
        else:
            return None


def _basis_functions_per_direction(element):
    """Number of basis functions per direction """
    from ufl import FiniteElement, TensorProductElement
    assert isinstance(element, (FiniteElement, TensorProductElement))
    degree = element.degree()
    if isinstance(degree, int):
        degree = (degree,) * world_dimension()

    basis_size = tuple(deg + 1 for deg in degree)

    # Anisotropic finite elements are not (yet) supported by Dune
    for size in basis_size:
        assert(size == basis_size[0])

    return basis_size


@kernel_cached
def pymbolic_coefficient_gradient(element, restriction, index, coeff_func, visitor_indices):
    sub_element = element
    grad_index = visitor_indices[0]
    if isinstance(element, MixedElement):
        sub_element = element.extract_component(index)[1]

    from ufl import FiniteElement
    assert isinstance(sub_element, (FiniteElement, TensorProductElement))

    # Number of basis functions per direction
    basis_size = _basis_functions_per_direction(sub_element)

    # Construct the matrix sequence for this sum factorization
    matrix_sequence = construct_basis_matrix_sequence(derivative=grad_index,
                                                      facedir=get_facedir(restriction),
                                                      facemod=get_facemod(restriction),
                                                      basis_size=basis_size,
                                                      )

    inp = LFSSumfactKernelInput(coeff_func=coeff_func,
                                element=element,
                                element_index=index,
                                restriction=restriction,
                                )

    # The sum factorization kernel object gathering all relevant information
    sf = SumfactKernel(matrix_sequence=matrix_sequence,
                       position_priority=grad_index,
                       input=inp,
                       )

    from dune.perftool.sumfact.vectorization import attach_vectorization_info
    vsf = attach_vectorization_info(sf)

    # If this sum factorization kernel was not used in the dry run we
    # just return 0
    if vsf == 0:
        return 0, None

    from dune.perftool.sumfact.realization import realize_sum_factorization_kernel
    var, insn_dep = realize_sum_factorization_kernel(vsf)

    return prim.Subscript(var, vsf.quadrature_index(sf)), None


@kernel_cached
def pymbolic_coefficient(element, restriction, index, coeff_func, visitor_indices):
    sub_element = element
    if isinstance(element, MixedElement):
        sub_element = element.extract_component(index)[1]
    from ufl import FiniteElement
    assert isinstance(sub_element, (FiniteElement, TensorProductElement))

    # Basis functions per direction
    basis_size = _basis_functions_per_direction(sub_element)

    # Construct the matrix sequence for this sum factorization
    matrix_sequence = construct_basis_matrix_sequence(facedir=get_facedir(restriction),
                                                      facemod=get_facemod(restriction),
                                                      basis_size=basis_size)

    inp = LFSSumfactKernelInput(coeff_func=coeff_func,
                                element=element,
                                element_index=index,
                                restriction=restriction,
                                )

    sf = SumfactKernel(matrix_sequence=matrix_sequence,
                       input=inp,
                       position_priority=3,
                       )

    from dune.perftool.sumfact.vectorization import attach_vectorization_info
    vsf = attach_vectorization_info(sf)

    # If this sum factorization kernel was not used in the dry run we
    # just return 0
    if vsf == 0:
        return 0, None

    # Add a sum factorization kernel that implements the evaluation of
    # the basis functions at quadrature points (stage 1)
    from dune.perftool.sumfact.realization import realize_sum_factorization_kernel
    var, _ = realize_sum_factorization_kernel(vsf)

    return prim.Subscript(var, vsf.quadrature_index(sf)), None


@iname
def sumfact_lfs_iname(element, bound, dim):
    assert(isinstance(bound, int))
    from dune.perftool.pdelab.driver import FEM_name_mangling
    name = "sumfac_lfs_{}_{}".format(FEM_name_mangling(element), dim)
    domain(name, bound)
    return name


@backend(interface="lfs_inames", name="sumfact")
def lfs_inames(element, restriction, number=1, context=''):
    from ufl import FiniteElement, TensorProductElement
    assert isinstance(element, (FiniteElement, TensorProductElement))
    if number == 0:
        return ()
    else:
        dim = world_dimension()
        return tuple(sumfact_lfs_iname(element, _basis_functions_per_direction(element)[d], d) for d in range(dim))


@backend(interface="evaluate_basis", name="sumfact")
@kernel_cached
def evaluate_basis(element, name, restriction):
    temporary_variable(name, shape=())
    quad_inames = quadrature_inames(element)
    inames = lfs_inames(element, restriction)
    facedir = get_facedir(restriction)

    # Collect the pairs of lfs/quad inames that are in use
    # On facets, the normal direction of the facet is excluded
    #
    # If facedir is not none, the length of inames and quad_inames is
    # different. For inames we want to skip the facedir direction, for
    # quad_inames we need all entries. Thats the reason for the
    # help_index.
    basis_per_dir = _basis_functions_per_direction(element)
    prod = ()
    help_index = 0
    for direction in range(len(inames)):
        if direction != facedir:
            prod = prod + (BasisTabulationMatrix(basis_size=basis_per_dir[direction]).pymbolic((prim.Variable(quad_inames[help_index]), prim.Variable(inames[direction]))),)
            help_index += 1

    # Add the missing direction on facedirs by evaluating at either 0 or 1
    if facedir is not None:
        facemod = get_facemod(restriction)
        prod = prod + (prim.Call(PolynomialLookup(name_polynomials(element.degree()), False),
                                 (prim.Variable(inames[facedir]), facemod)),)

    # Issue the product
    instruction(expression=prim.Product(prod),
                assignee=prim.Variable(name),
                forced_iname_deps=frozenset(quad_inames + inames),
                forced_iname_deps_is_final=True,
                )


def pymbolic_basis(element, restriction, number):
    # If this is a test function we omit it!
    if number == 0:
        return 1

    assert not isinstance(element, MixedElement)

    name = "phi_{}".format(FEM_name_mangling(element))
    name = restricted_name(name, restriction)
    evaluate_basis(element, name, restriction)

    return prim.Variable(name)


@backend(interface="evaluate_grad", name="sumfact")
@kernel_cached
def evaluate_reference_gradient(element, name, restriction, index):
    dim = world_dimension()
    temporary_variable(name, shape=())
    quad_inames = quadrature_inames(element)
    inames = lfs_inames(element, restriction)
    facedir = get_facedir(restriction)

    # Map the direction to a quadrature iname
    quadinamemapping = {}
    i = 0
    for d in range(local_dimension()):
        if d == facedir:
            i = i + 1
        quadinamemapping[i] = quad_inames[d]
        i = i + 1

    prod = []
    for i in range(dim):
        if i != facedir:
            tab = BasisTabulationMatrix(basis_size=_basis_functions_per_direction(element)[i],
                                        derivative=index == i)
            prod.append(tab.pymbolic((prim.Variable(quadinamemapping[i]), prim.Variable(inames[i]))))

    if facedir is not None:
        facemod = get_facemod(restriction)
        prod.append(prim.Call(PolynomialLookup(name_polynomials(element.degree()), index == facedir),
                              (prim.Variable(inames[facedir]), facemod)),)

    instruction(assignee=prim.Variable(name),
                expression=prim.Product(tuple(prod)),
                forced_iname_deps=frozenset(quad_inames + inames),
                forced_iname_deps_is_final=True,
                )


def pymbolic_reference_gradient(element, restriction, number, indices):
    # If this is a test function, we omit it.
    if number == 0:
        return 1, None

    assert len(indices) == 1
    index, = indices

    # TODO: Change name?
    name = "js_{}".format(FEM_name_mangling(element))
    name = restricted_name(name, restriction)
    name = "{}_{}".format(name, index)
    evaluate_reference_gradient(element, name, restriction, index)

    return prim.Variable(name), None
